/* -*- mode: c; coding: koi8-r -*- */
/* $Id: nls_utf8.c,v 1.2 2002/12/01 23:04:59 cher Exp $ */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "nls.h"

#include <stdio.h>

static void
uni2char(unsigned char ch, unsigned char cl,
         unsigned char *out, int boundlen, int *outlen)
{
  unsigned int ucs2 = (ch << 8) | cl;

  if (boundlen <= 0)
    return;

  if (ucs2 <= 0x7f) {
    out[0] = ucs2;
    *outlen = 1;
  } else if (ucs2 <= 0x7ff) {
    if (boundlen > 1) {
      out[0] = (ucs2 >> 6) | 0xc0;
      out[1] = (ucs2 & 0x3f) | 0x80;
    }
    *outlen = 2;
  } else {
    if (boundlen > 2) {
      out[0] = (ucs2 >> 12) | 0xe0;
      out[1] = ((ucs2 >> 6) & 0x3f) | 0x80;
      out[2] = (ucs2 & 0x3f) | 0x80;
    }
    *outlen = 3;
  }
}

static int
unilen(unsigned char ch, unsigned char cl)
{
  unsigned int ucs2 = (ch << 8) | cl;
  if (ucs2 <= 0x7f) return 1;
  if (ucs2 <= 0x7ff) return 2;
  return 3;
}

static void
char2uni(const unsigned char *rawstring, int *offset,
         unsigned char *uni1, unsigned char *uni2)
{
  if (!(*rawstring & 0x80)) {
    *uni1 = *rawstring;
    *uni2 = 0;
    *offset = 1;
  } else if ((*rawstring & 0xe0) == 0xc0) {
    if ((rawstring[1] & 0xc0) == 0x80) {
      *uni1 = (rawstring[1] & 0x3f) | ((rawstring[0] & 0x03) << 6);
      *uni2 = (rawstring[0] & 0x1c) >> 2;
      *offset = 2;
    } else if (!rawstring[1]) {
      *uni1 = '?';
      *uni2 = 0;
      *offset = 1;
    } else {
      *uni1 = '?';
      *uni2 = 0;
      *offset = 2;
    }
  } else if ((*rawstring & 0xf0) == 0xe0) {
    if ((rawstring[1] & 0xc0) == 0x80 && (rawstring[2] & 0xc0) == 0x80) {
      *uni1 = (rawstring[2] & 0x3f) | ((rawstring[1] & 0x03) << 6);
      *uni2 = ((rawstring[0] & 0x0f) << 4) | ((rawstring[1] & 0x3c) >> 2);
      *offset = 3;
    } if (!rawstring[1]) {
      *uni1 = '?';
      *uni2 = 0;
      *offset = 1;
    } if (!rawstring[2]) {
      *uni1 = '?';
      *uni2 = 0;
      *offset = 2;
    } else {
      *uni1 = '?';
      *uni2 = 0;
      *offset = 3;
    }
  } else {
    *uni1 = '?';
    *uni2 = 0;
    *offset = 1;
  }
}

static struct nls_table table =
{
  "utf-8",
  0,
  uni2char,
  char2uni,
  unilen
};

struct nls_table *nls_table_utf8 = &table;
