/* -*- mode: c -*- */
/* $Id: dosexec.c,v 1.1 2001/12/02 09:11:07 cher Exp $ */

/* Copyright (C) 2001 Alexander Chernov <cher@ispras.ru> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* This program is intended to be compiled with Borland C 3.1
 * in MS-DOS environment. It redirect the standard input from
 * file "./input.", standard output to "./output.",
 * standard error to "./error.". Then it starts "./program.exe"
 * program. The return code of the program then is written to
 * "./e.".
 *
 * The program should be compiled for "small" memory model
 * (or even better "tiny" one).
 * FIXME: Program code occupies about 20KB, program data is no more,
 * than 10KB, but Borland C allocates the whole memory segment (64KB)
 * for data. If it is possible to optimize data usage, the program
 * whould occupy only 25-30KB in memory instead of the current 90Kb.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <io.h>
#include <errno.h>
#include <process.h>

#define DOSEXEC_ERROR 100

void do_exit(int code)
{
  FILE *f = NULL;

  if (!(f = fopen("./e.", "wb"))) {
    fprintf(stderr, "dosexec: cannot open ./e for writing");
    exit(DOSEXEC_ERROR);
  }
  fprintf(f, "%d\r\n", code);
  if (ferror(f)) {
    fprintf(stderr, "dosexec: i/o error on ./e");
    exit(DOSEXEC_ERROR);
  }
  if (fclose(f) < 0) {
    fprintf(stderr, "dosexec: i/o error on ./e");
    exit(DOSEXEC_ERROR);
  }
  exit(code);
}

int checked_open(char const *path)
{
  int fd = -1;
  if ((fd = _open(path, O_RDONLY)) < 0) {
    fprintf(stderr, "dosexec: open(%s,O_RDONLY) failed: %s\n",
	    path, strerror(errno));
    do_exit(DOSEXEC_ERROR);
  }
  return fd;
}

int checked_creat(char const *path)
{
  int fd = -1;
  if ((fd = _creat(path, 0)) < 0) {
    fprintf(stderr, "dosexec: creat(%s,0) failed: %s\n",
	    path, strerror(errno));
    do_exit(DOSEXEC_ERROR);
  }
  return fd;
}

void checked_dup2(int fdold, int fdnew)
{
  if (dup2(fdold, fdnew) < 0) {
    fprintf(stderr, "dosexec: dup2(%d,%d) failed: %s\n",
	    fdold, fdnew, strerror(errno));
    do_exit(DOSEXEC_ERROR);
  }
}

void checked_close(int fd)
{
  if (_close(fd) < 0) {
    fprintf(stderr, "dosexec: close(%d) failed: %s\n",
	    fd, strerror(errno));
    do_exit(DOSEXEC_ERROR);
  }
}

int main(void)
{
  int fdout = -1, fdin = -1, fderr = -1;
  int stat;

  // redirect stardard streams
  fdin  = checked_open("./input.");
  fdout = checked_creat("./output.");
  fderr = checked_creat("./error.");
  checked_dup2(fdin, 0);
  checked_dup2(fdout, 1);
  checked_dup2(fderr, 2);
  checked_close(fdin);
  checked_close(fdout);
  checked_close(fderr);
  // run program.exe
  if ((stat = spawnl(P_WAIT, "./program.exe", "./program.exe", NULL)) < 0) {
    fprintf(stderr, "dosexec: spawnl(...) failed: %s\n",
	    strerror(errno));
    do_exit(DOSEXEC_ERROR);
  }
  // copy its return code
  do_exit(stat);
  return 0;
}
