/* -*- mode: c -*- */
/* $Id: filter_test.c,v 1.3 2002/12/13 19:38:33 cher Exp $ */

/* Copyright (C) 2002 Alexander Chernov <cher@ispras.ru> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "filter_tree.h"
#include "filter_eval.h"
#include "pathutl.h"
#include "prepare.h"
#include "teamdb.h"
#include "runlog.h"

#include <reuse/logger.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>

static unsigned char buf[1024];
extern int filter_expr_nerrs;
extern struct filter_tree *filter_expr_lval;

void
filter_runs(struct filter_tree *expr)
{
  struct filter_env env;
  int i, r;
  unsigned char resbuf[32];

  memset(&env, 0, sizeof(env));
  env.mem = filter_tree_new();
  env.maxlang = max_lang;
  env.langs = langs;
  env.maxprob = max_prob;
  env.probs = probs;
  env.rtotal = run_get_total();
  run_get_header(&env.rhead);
  env.rentries = alloca(env.rtotal * sizeof(env.rentries[0]));
  env.cur_time = time(0);
  run_get_all_entries(env.rentries);

  for (i = env.rtotal - 1; i >= 0; i--) {
    env.rid = i;
    if (expr) {
      r = filter_tree_bool_eval(&env, expr);
      if (r < 0) {
        fprintf(stderr, "evaluation of run %d: %s\n", i, filter_strerror(-r));
        continue;
      }
      if (!r) continue;
    }
    filter_tree_result_str(resbuf,sizeof(resbuf),env.rentries[i].status);
    printf("%-4d %-6zu %-16s %-4d %-8.8s %-2.2s %-5.5s %-3.3s %-4d %d\n",
           i, env.rentries[i].size,
           run_unparse_ip(env.rentries[i].ip),
	   env.rentries[i].team,
           teamdb_get_login(env.rentries[i].team),
           env.probs[env.rentries[i].problem]->short_name,
           env.langs[env.rentries[i].language]->short_name,
           resbuf, env.rentries[i].test,
           env.rentries[i].score);
    filter_tree_clear(env.mem);
  }

  filter_tree_delete(env.mem);
}

int main(int argc, char **argv)
{
  path_t  cpp_opts = { 0 };
  int     code = 0;
  int     p_flags = 0, T_flag = 0;
  int     i = 1;
  struct filter_tree_mem *mem;

  logger_set_level(-1, LOG_WARNING);

  if (argc == 1) goto print_usage;
  code = 1;

  while (i < argc) {
    if (!strcmp(argv[i], "-T")) {
      i++;
      T_flag = 1;
    } else if (!strncmp(argv[i], "-D", 2)) {
      if (cpp_opts[0]) pathcat(cpp_opts, " ");
      pathcat(cpp_opts, argv[i++]);
    } else if (!strcmp(argv[i], "-E")) {
      i++;
      p_flags |= PREPARE_USE_CPP;
    } else break;
  }
  if (i >= argc) goto print_usage;

  if (prepare(argv[i], p_flags, PREPARE_SERVE, cpp_opts) < 0) return 1;

  if (T_flag) {
    print_configuration(stdout);
    return 0;
  }
  if (global->contest_id <= 0) {
    err("contest_id is not defined");
    return 1;
  }
  if (teamdb_open_client(global->socket_path, global->contest_id) < 0)
    return 1;
  if (run_open(global->run_log_file, RUN_LOG_READONLY) < 0) return 1;

  while (fgets(buf, sizeof(buf), stdin)) {
    mem = filter_tree_new();
    filter_expr_set_string(buf, mem, 0);
    filter_expr_init_parser(mem, 0);
    i = filter_expr_parse();
    if (i + filter_expr_nerrs > 0) continue;
    if (filter_expr_lval) {
      if (filter_expr_lval->type != FILTER_TYPE_BOOL) {
        fprintf(stderr, "bool expression expected\n");
        continue;
      }
      filter_tree_print(filter_expr_lval, stdout, "");
    }
    filter_runs(filter_expr_lval);
    filter_tree_delete(mem);
  }

  return 0;

 print_usage:
  printf("Usage: %s [ OPTS ] config-file\n", argv[0]);
  printf("  -T     - print configuration and exit\n");
  printf("  -E     - enable C preprocessor\n");
  printf("  -DDEF  - define a symbol for preprocessor\n");
  return code;
}

/**
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE" "va_list" "jmp_buf")
 *  eval: (set-language-environment "Cyrillic-KOI8")
 * End:
 */
