/* -*- c -*- */
/* $Id: localdb.c,v 1.3 2002/03/22 20:36:42 cher Exp $ */

/* Copyright (C) 2001,2002 Alexander Chernov <cher@ispras.ru> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "localdb.h"

#include <reuse/xalloc.h>

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

static void err(char const *, ...);

enum
{
  LOGIN_FIELD_LEN = 32,
  ID_FIELD_LEN = 32,
  FLAGS_FIELD_LEN = 32,
  TNAME_FIELD_LEN = 64,
  INST_FULL_FIELD_LEN = 256,
  INST_SHORT_FIELD_LEN = 32,
  FAC_FULL_FIELD_LEN = 256,
  FAC_SHORT_FIELD_LEN = 32,
  NAME1_FIELD_LEN = 128,
  COURSE1_FIELD_LEN = 32,
  NAME2_FIELD_LEN = 128,
  COURSE2_FIELD_LEN = 32,
  NAME3_FIELD_LEN = 128,
  COURSE3_FIELD_LEN = 32,
  NAME4_FIELD_LEN = 128,
  COURSE4_FIELD_LEN = 32,
  NAME_ADV_FIELD_LEN = 128,
  POST_ADV_FIELD_LEN = 128,
  NAME_COACH_FIELD_LEN = 128,
  POST_COACH_FIELD_LEN = 128,
  EMAIL_FIELD_LEN = 128,
  PHONE_FIELD_LEN = 128,
  HOMEPAGE_FIELD_LEN = 128
};

struct localdb_elem
{
  char *login;
  int  id;
  char *flags;
  char *tname;
  char *inst_full;
  char *inst_short;
  char *fac_full;
  char *fac_short;
  char *name1;
  char *course1;
  char *name2;
  char *course2;
  char *name3;
  char *course3;
  char *name4;
  char *course4;
  char *name_adv;
  char *post_adv;
  char *name_coach;
  char *post_coach;
  char *email;
  char *phone;
  char *homepage;
};

struct localdb_elem_arr
{
  int                   a, u;
  struct localdb_elem **v;
};

struct localdb
{
  char *top_comments;
  char *bottom_comments;
  struct localdb_elem_arr items;
  int   id_size;
  int  *id_map;
};

static char b_login[LOGIN_FIELD_LEN];
static char b_id[ID_FIELD_LEN];
static char b_flags[FLAGS_FIELD_LEN];
static char b_tname[TNAME_FIELD_LEN];
static char b_inst_full[INST_FULL_FIELD_LEN];
static char b_inst_short[INST_SHORT_FIELD_LEN];
static char b_fac_full[FAC_FULL_FIELD_LEN];
static char b_fac_short[FAC_SHORT_FIELD_LEN];
static char b_name1[NAME1_FIELD_LEN];
static char b_course1[COURSE1_FIELD_LEN];
static char b_name2[NAME1_FIELD_LEN];
static char b_course2[COURSE1_FIELD_LEN];
static char b_name3[NAME1_FIELD_LEN];
static char b_course3[COURSE1_FIELD_LEN];
static char b_name4[NAME1_FIELD_LEN];
static char b_course4[COURSE1_FIELD_LEN];
static char b_name_adv[NAME_ADV_FIELD_LEN];
static char b_post_adv[POST_ADV_FIELD_LEN];
static char b_name_coach[NAME_COACH_FIELD_LEN];
static char b_post_coach[POST_COACH_FIELD_LEN];
static char b_email[EMAIL_FIELD_LEN];
static char b_phone[PHONE_FIELD_LEN];
static char b_homepage[HOMEPAGE_FIELD_LEN];

struct fieldinfo
{
  char *name;
  int   maxlen;
  char *ptr;
};

static struct fieldinfo localdb_fieldinfo [] =
{
  { "login", LOGIN_FIELD_LEN, b_login },
  { "id", ID_FIELD_LEN, b_id },
  { "flags", FLAGS_FIELD_LEN, b_flags },
  { "tname", TNAME_FIELD_LEN, b_tname },
  { "inst_full", INST_FULL_FIELD_LEN, b_inst_full },
  { "inst_short", INST_SHORT_FIELD_LEN, b_inst_short },
  { "fac_full", FAC_FULL_FIELD_LEN, b_fac_full },
  { "fac_short", FAC_SHORT_FIELD_LEN, b_fac_short },
  { "name1", NAME1_FIELD_LEN, b_name1 },
  { "course1", COURSE1_FIELD_LEN, b_course1 },
  { "name2", NAME2_FIELD_LEN, b_name2 },
  { "course2", COURSE2_FIELD_LEN, b_course2 },
  { "name3", NAME3_FIELD_LEN, b_name3 },
  { "course3", COURSE3_FIELD_LEN, b_course3 },
  { "name4", NAME4_FIELD_LEN, b_name4 },
  { "course4", COURSE4_FIELD_LEN, b_course4 },
  { "name_adv", NAME_ADV_FIELD_LEN, b_name_adv },
  { "post_adv", POST_ADV_FIELD_LEN, b_post_adv },
  { "name_coach", NAME_COACH_FIELD_LEN, b_name_coach },
  { "post_coach", POST_COACH_FIELD_LEN, b_post_coach },
  { "email", EMAIL_FIELD_LEN, b_email },
  { "phone", PHONE_FIELD_LEN, b_phone },
  { "homepage", HOMEPAGE_FIELD_LEN, b_homepage },

  { 0, 0, 0 }
};

static int
read_fields(char const *s, int separator, const struct fieldinfo *fi,
            char const *file, int line)
{
  int i, j;
  char const *p = s;

  for (i = 0; fi[i].maxlen && *p; i++) {
    j = 0;
    memset(fi[i].ptr, 0, fi[i].maxlen);
    while (*p != 0 && *p != separator && *p != '\n') {
      if (j >= fi[i].maxlen - 1) {
        err("%s: %d: read_fields: field %s is too long", file,
            line, fi[i].name);
        return -1;
      }
      fi[i].ptr[j++] = *p++;
    }
    if (*p == separator || *p == '\n') p++;
  }
  if (fi[i].maxlen != 0) {
    err("%s: %d: read_fields: too few fields read: %d", file, line, i);
    return -1;
  }
  return i;
}

void
free_localdb_elem(struct localdb_elem *pdbe)
{
  if (!pdbe) return;
  xfree(pdbe->login);
  xfree(pdbe->flags);
  xfree(pdbe->tname);
  xfree(pdbe->inst_full);
  xfree(pdbe->inst_short);
  xfree(pdbe->fac_full);
  xfree(pdbe->fac_short);
  xfree(pdbe->name1);
  xfree(pdbe->course1);
  xfree(pdbe->name2);
  xfree(pdbe->course2);
  xfree(pdbe->name3);
  xfree(pdbe->course3);
  xfree(pdbe->name4);
  xfree(pdbe->course4);
  xfree(pdbe->name_adv);
  xfree(pdbe->post_adv);
  xfree(pdbe->name_coach);
  xfree(pdbe->post_coach);
  xfree(pdbe->email);
  xfree(pdbe->phone);
  xfree(pdbe->homepage);
  XMEMZERO(pdbe, 1);
  xfree(pdbe);
}

void
free_localdb(struct localdb *pdb)
{
  int i;

  xfree(pdb->top_comments);
  xfree(pdb->bottom_comments);
  for (i = 0; i < pdb->items.u; i++) {
    free_localdb_elem(pdb->items.v[i]);
  }
  if (pdb->items.a > 0) {
    XMEMZERO(pdb->items.v, pdb->items.a);
  }
  xfree(pdb->items.v);
  xfree(pdb->id_map);
  XMEMZERO(pdb, 1);  
}

struct localdb *
read_localdb(char const *name)
{
  struct localdb *pdb = 0;
  struct localdb_elem *pelem = 0;
  FILE *in = 0;
  int line = 0;
  int len;
  int tid, n, i;
  int state = 0;
  char linebuf[1024];

  if (!(in = fopen(name, "r"))) {
    err("read_localdb: cannot open `%s'", name);
    goto cleanup;
  }
  XCALLOC(pdb, 1);

  while (fgets(linebuf, sizeof(linebuf), in) != NULL) {
    line++;
    len = strlen(linebuf);
    if (len == sizeof(linebuf) - 1 && linebuf[len - 1] != '\n') {
      err("%s: %d: read_localdb: line is too long", name, line);
      goto cleanup;
    }
    if (linebuf[0] == '#') {
      if (state == 0) {
        pdb->top_comments = xstrmerge1(pdb->top_comments, linebuf);
      } else if (state == 2) {
        pdb->bottom_comments = xstrmerge1(pdb->bottom_comments, linebuf);
      }
      if (!strcmp(linebuf, "###\n")) {
        state++;
        if (state == 2)
          pdb->bottom_comments = xstrmerge1(pdb->bottom_comments, linebuf);
      }
      continue;
    }
    if (read_fields(linebuf, ';', localdb_fieldinfo, name, line) < 0)
      goto cleanup;
    if (sscanf(b_id, "%d %n", &tid, &n) != 1 || b_id[n]
        || tid <= 0 || tid > 65535) {
      err("%s: %d: invalid team_id: %s, %d", name, line, b_id, tid);
      goto cleanup;
    }

    XCALLOC(pelem, 1);
    pelem->login = xstrdup(b_login);
    pelem->id = tid;
    pelem->flags = xstrdup(b_flags);
    pelem->tname = xstrdup(b_tname);
    pelem->inst_full = xstrdup(b_inst_full);
    pelem->inst_short = xstrdup(b_inst_short);
    pelem->fac_full = xstrdup(b_fac_full);
    pelem->fac_short = xstrdup(b_fac_short);
    pelem->name1 = xstrdup(b_name1);
    pelem->course1 = xstrdup(b_course1);
    pelem->name2 = xstrdup(b_name2);
    pelem->course2 = xstrdup(b_course2);
    pelem->name3 = xstrdup(b_name3);
    pelem->course3 = xstrdup(b_course3);
    pelem->name4 = xstrdup(b_name4);
    pelem->course4 = xstrdup(b_course4);
    pelem->name_adv = xstrdup(b_name_adv);
    pelem->post_adv = xstrdup(b_post_adv);
    pelem->name_coach = xstrdup(b_name_coach);
    pelem->post_coach = xstrdup(b_post_coach);
    pelem->email = xstrdup(b_email);
    pelem->phone = xstrdup(b_phone);
    pelem->homepage = xstrdup(b_homepage);
    XEXPAND2(pdb->items);
    pdb->items.v[pdb->items.u++] = pelem;
    pelem = 0;
  }

  fclose(in);

  pdb->id_size = -1;
  for (i = 0; i < pdb->items.u; i++) {
    if (pdb->items.v[i]->id > pdb->id_size)
      pdb->id_size = pdb->items.v[i]->id;
  }
  pdb->id_size++;
  if (pdb->id_size > 0) {
    XCALLOC(pdb->id_map, pdb->id_size);
    for (i = 0; i < pdb->id_size; i++)
      pdb->id_map[i] = -1;
  }
  for (i = 0; i < pdb->items.u; i++) {
    pdb->id_map[pdb->items.v[i]->id] = i;
  }

  return pdb;

 cleanup:
  if (in) fclose(in);
  free_localdb(pdb);
  free_localdb_elem(pelem);
  return 0;
}

int
localdb_id_size(struct localdb *db)
{
  return db->id_size;
}

int
localdb_get_index(struct localdb *db, int teamid)
{
  return db->id_map[teamid];
}

char *
localdb_get_name(struct localdb *db, int teamid)
{
  return db->items.v[db->id_map[teamid]]->tname;
}

char *
localdb_get_login(struct localdb *db, int teamid)
{
  return db->items.v[db->id_map[teamid]]->login;
}

int
localdb_training_team(struct localdb *db, int teamid)
{
  char *flags = db->items.v[db->id_map[teamid]]->flags;
  if (strchr(flags, 't')) return 1;
  return 0;
}

int
localdb_selection_team(struct localdb *db, int teamid)
{
  char *flags = db->items.v[db->id_map[teamid]]->flags;
  if (!strchr(flags, 't')) return 1;
  return 0;
}

int
localdb_final_team(struct localdb *db, int teamid)
{
  char *flags = db->items.v[db->id_map[teamid]]->flags;
  if (strchr(flags, 'f')) return 1;
  return 0;
}

#if 0
int
main(int argc, char *argv[])
{
  struct localdb      *pdb;
  struct localdb_elem *pel;
  int i;

  if (argc != 2) {
    err("wrong number of arguments");
    return 1;
  }
  if (!(pdb = read_localdb(argv[1]))) return 1;

  printf("%s", pdb->top_comments);
  for (i = 0; i < pdb->items.u; i++) {
    pel = pdb->items.v[i];

    printf("%s;%d;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s\n",
           pel->login, pel->id, pel->flags, pel->tname,
           pel->inst_full, pel->inst_short,
           pel->fac_full, pel->fac_short,
           pel->name1, pel->course1,
           pel->name2, pel->course2,
           pel->name3, pel->course3,
           pel->name4, pel->course4,
           pel->name_adv, pel->post_adv,
           pel->name_coach, pel->post_coach,
           pel->email, pel->phone, pel->homepage);
  }
  printf("%s", pdb->bottom_comments);

  return 0;
}
#endif

static void
err(char const *format, ...)
{
  va_list args;

  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
  putc('\n', stderr);
}

/**
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE")
 * End:
 */

