/* -*- c -*- */
/* $Id: make-teamdb.c,v 1.3 2002/09/01 06:13:26 cher Exp $ */

/* Copyright (C) 2001,2002 Alexander Chernov <cher@ispras.ru> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "idmap.h"
#include "localdb.h"

#include <stdio.h>
#include <string.h>

enum
{
  TRAINING_TOUR,
  SELECTION_TOUR,
  FINAL_TOUR
};

int
main(int argc, char *argv[])
{
  struct localdb *db1;
  struct idmap   *db2;
  int dbsize;
  int tour;
  int i, idb1, idb2;
  char *login;
  char *name;
  char *login1;
  char *name2;

  if (argc != 4) {
    printf("Usage: make-teamdb <localdb> <idmap> [t|s|f]\n");
    return 0;
  }
  if (!strcmp(argv[3], "t")) {
    tour = TRAINING_TOUR;
  } else if (!strcmp(argv[3], "s")) {
    tour = SELECTION_TOUR;
  } else if (!strcmp(argv[3], "f")) {
    tour = FINAL_TOUR;
  } else {
    fprintf(stderr, "invalid tour\n");
    return 1;
  }

  if (!(db1 = read_localdb(argv[1]))) return 1;
  if (!(db2 = read_idmap(argv[2]))) return 1;

  // check consistency
  dbsize = localdb_id_size(db1);
  if (dbsize != idmap_id_size(db2)) {
    fprintf(stderr, "base sizes mismatch\n");
    return 1;
  }
  for (i = 0; i < dbsize; i++) {
    idb1 = localdb_get_index(db1, i);
    idb2 = idmap_get_index(db2, i);
    if (idb1 >= 0 && idb2 >= 0 && i == 0) {
      fprintf(stderr, "teamid == 0 used\n");
      return 1;
    }
    if (idb1 == -1 && idb2 == -1) continue;
    if (idb1 != -1 && idb2 != -1) continue;
    fprintf(stderr, "teamid %d mismatch\n", i);
  }
  for (i = 0; i < dbsize; i++) {
    if (localdb_get_index(db1, i) < 0) continue;
    name = localdb_get_name(db1, i);
    name2 = idmap_get_name(db2, i);
    if (strcmp(name, name2)) {
      fprintf(stderr, "names for %d does not match\n", i);
      return 1;
    }
    login = localdb_get_login(db1, i);
    login1 = idmap_get_login(db2, i);
    if (strcmp(login, login1)) {
      fprintf(stderr, "logins for %d does not match\n", i);
      return 1;
    }
  }

  for (i = 0; i < dbsize; i++) {
    if (localdb_get_index(db1, i) < 0) continue;
    login1 = 0;
    switch (tour) {
    case TRAINING_TOUR:
      if (localdb_training_team(db1, i))
        login1 = idmap_get_login1(db2, i);
      break;
    case SELECTION_TOUR:
      if (localdb_selection_team(db1, i))
        login1 = idmap_get_login1(db2, i);
      break;
    case FINAL_TOUR:
      if (localdb_final_team(db1, i))
        login1 = idmap_get_login2(db2, i);
      break;
    default:
      fprintf(stderr, "bad tour %d\n", tour);
      return 1;
    }
    if (!login1) continue;
    if (!*login1) login1 = localdb_get_login(db1, i);
    name = localdb_get_name(db1, i);
    printf("%s:%d::%s\n", login1, i, name);
  }

  return 0;
}

/**
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE")
 * End:
 */

