/* -*- c -*- */
/* $Id: mkpasswd.c,v 1.10 2002/10/15 21:45:32 cher Exp $ */

/* Copyright (C) 2000-2002 Alexander Chernov <cher@ispras.ru> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "teamdb.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#if CONF_HAS_LIBINTL - 0 == 1
#include <libintl.h>
#define _(x) gettext(x)
#else
#define _(x) x
#endif

static char const allowed_chars[]=
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

void
generate_random_passwd(char *buf, int size)
{
  int          i;
  unsigned int val;

  for (i = 0; i < size; i++) {
    val = random();
    val = (val >> 9) & 0x3F;
    buf[i] = allowed_chars[val];
  }
  buf[i] = 0;
}

int
main(int argc, char *argv[])
{
  int i, m;
  char newpasswd[16];
  char passwd2[24];
  int  tex_flag = 0;

  if (argc < 3 || argc > 5) {
    printf("  Usage: %s <teamdb file> <passwd file> [ random ]\n",
           argv[0]);
    return 0;
  }

  if (teamdb_open(argv[1], argv[2], 1) < 0) return 1;

  if (argc == 5) {
    int team_id = teamdb_lookup_login(argv[3]);
    char passwd_raw[64];
    char passwd_base64[64];

    if (!team_id) {
      fprintf(stderr, "%s: team %s does not exist\n",
              argv[0], argv[3]);
      return 1;
    }
    if (strlen(argv[4]) > 16) {
      fprintf(stderr, "%s: password too long\n", argv[0]);
      return 1;
    }
    strcpy(passwd_raw, argv[4]);
    teamdb_scramble_passwd(passwd_raw, passwd_base64);
    teamdb_set_scrambled_passwd(team_id, passwd_base64);
  }

  if (argc == 4 && !strcmp(argv[3], "random-tex")) {
    tex_flag = 1;
    argv[3] = "random";
  }

  if (argc == 4 && !strcmp(argv[3], "random")) {
    if (tex_flag) {
      printf("%s",
             "% -*- mode: LaTeX; coding: koi8-r -*-\n"
             "\n"
             "\\makeatletter\n"
             "\\nonstopmode\n"
             "\\documentclass[12pt]{article}\n"
             "\\usepackage[koi8-r]{inputenc}\n"
             "\\usepackage[english,russian]{babel}\n"
             "\\usepackage[a4paper]{geometry}\n"
             "\\usepackage{amssymb}\n"
             "\\usepackage{indentfirst}\n"
             "\\usepackage{latexsym}\n"
             "\\usepackage{pscyr}\n"
             "\\usepackage{mathptm}\n"
             "\\renewcommand{\\rmdefault}{ftm}\n"
             "\\renewcommand{\\ttdefault}{fcr}\n"
             "\n"
             "\\begin{document}\n");
    }
    srandom(time(0));
    m = teamdb_get_max_team_id();
    for (i = 1; i <= m; i++) {
      if (!teamdb_lookup(i)) continue;
      generate_random_passwd(newpasswd, 9);
      if (tex_flag) {
        printf("\\begin{verbatim}\n"
               "Team name:  \"%s\"\n"
               "Team login: %s\n"
               "Password:   %s\n\n"
               "\\end{verbatim}\n"
               "\\hrule\n",
               teamdb_get_name(i), teamdb_get_login(i), newpasswd);
      } else {
        printf("Team name:  \"%s\"\n"
               "Team login: %s\n"
               "Password:   %s\n\n",
               teamdb_get_name(i),
               teamdb_get_login(i), newpasswd);
      }
      teamdb_scramble_passwd(newpasswd, passwd2);
      teamdb_set_scrambled_passwd(i, passwd2);
    }
    if (tex_flag) {
      printf("%s",
             "\\end{document}\n"
             "\n"
             "% Local variables:\n"
             "%  eval: (set-language-environment \"Cyrillic-KOI8\")\n"
             "% End:\n");

    }
  }

  if (teamdb_write_passwd(argv[2]) < 0) return 1;

  return 0;
}

/**
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE")
 * End:
 */
