/* -*- c -*- */
/* $Id: send-passwords.c,v 1.2 2001/10/05 22:54:19 cher Exp $ */

/* Copyright (C) 2001 Alexander Chernov <cher@ispras.ru> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "inetdb.h"
#include "teamdb.h"
#include "fileutl.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static int
sformat_message(char *buf, size_t maxsize, char const *format,
                char const *name, char const *login, char const *passwd,
                char const *email)
{
  char *s = buf;
  char const *q;
  char const *p = format;
  int ws = 0;

  /* actual mode */
  if (!buf || !maxsize) {
  } else if (maxsize == 1) {
    buf[0] = 0;
  } else {
    while (ws < maxsize - 1 && *p) {
      if (*p == '%') {
        q = 0;
        switch (p[1]) {
        case 'm':               /* email */
          q = email;
          p += 2;
          break;
        case 'l':               /* login */
          q = login;
          p += 2;
          break;
        case 'p':               /* password */
          q = passwd;
          p += 2;
          break;
        case 'n':               /* name */
          q = name;
          p += 2;
          break;
        default:
          *s++ = *p++;
          ws++;
        }
        if (q) {
          while (ws < maxsize - 1 && *q) {
            *s++ = *q++;
            ws++;
          }
          if (*q) {
            ws++;
            q++;
          }
        }
      } else {
        *s++ = *p++;
        ws++;
      }
    }
    *s = 0;
  }

  /* dummy mode */
  while (*p) {
    if (*p == '%') {
      q = 0;
      switch (p[1]) {
      case 'm':               /* email */
        q = email;
        p += 2;
        break;
      case 'l':               /* login */
        q = login;
        p += 2;
        break;
      case 'p':               /* password */
        q = passwd;
        p += 2;
        break;
      case 'n':               /* name */
        q = name;
        p += 2;
        break;
      default:
        p++;
        ws++;
      }
      if (q) ws += strlen(q);
    } else {
      p++;
      ws++;
    }
  }

  return ws;
}

int
main(int argc, char *argv[])
{
  struct inetdb *idb;
  int dbsize, i, j, n;
  char *template = 0;
  size_t templ_size = 0;
  char *login;
  char *name;
  char *emails;
  char  email[128];
  char  password[64];
  char  message[16384];
  int   msg_size;
  FILE *out = 0;

  if (argc != 6) {
    printf("Usage: send-passwords <inetdb> <teamdb> <passwd> <template> <sendprog>\n");
    return 0;
  }

  if (!(idb = read_inetdb(argv[1]))) return 1;
  if (teamdb_open(argv[2], argv[3], 0) < 0) return 1;
  if (generic_read_file(&template, 0, &templ_size, 0, 0, argv[4], 0) < 0)
    return 1;
  dbsize = inetdb_id_size(idb);

  for (i = 0; i < dbsize; i++) {
    if (inetdb_get_index(idb, i) < 0) continue;
    login = inetdb_get_login(idb, i);
    name = inetdb_get_name(idb, i);
    emails = inetdb_get_email(idb, i);
    if (teamdb_get_plain_password(i, password, sizeof(password)) < 0)
      return 1;

    j = 0;
    while (sscanf(emails + j, "%s%n", email, &n) == 1) {
      j += n;

      msg_size = sformat_message(message, sizeof(message),
                                 template, name, login, password, email);
      if (msg_size >= sizeof(message)) {
        fprintf(stderr, "message buffer is too small\n");
        return 1;
      }
      if (!(out = popen(argv[5], "w"))) {
        fprintf(stderr, "cannot run program %s\n", argv[5]);
        return 1;
      }
      fprintf(out, "%s", message);
      fflush(out);
      if (ferror(out)) {
        fprintf(stderr, "write error to program %s\n", argv[5]);
        return 1;
      }
      pclose(out);
      out = 0;
    }
  }

  return 0;
}

/**
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE")
 * End:
 */

