/* -*- c -*- */
/* $Id: submit.c,v 1.10 2002/10/09 05:31:08 cher Exp $ */

/* Copyright (C) 2000-2002 Alexander Chernov <cher@ispras.ru> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pathutl.h"
#include "prepare.h"
#include "teamdb.h"
#include "fileutl.h"

#include <reuse/logger.h>

#include <stdio.h>
#include <string.h>

#if CONF_HAS_LIBINTL - 0 == 1
#include <libintl.h>
#define _(x) gettext(x)
#else
#define _(x) x
#endif

int
main(int argc, char *argv[])
{
  int team;
  int prob;
  int lang;
  int n;
  int code = 0, i = 1, p_flags = PREPARE_QUIET;
  path_t cpp_opts;
  path_t uniq;
  char   cmd[64];

  if (argc == 1) goto print_usage;
  code = 1;

  while (i < argc) {
    if (!strncmp(argv[i], "-D", 2)) {
      if (cpp_opts[0]) pathcat(cpp_opts, " ");
      pathcat(cpp_opts, argv[i++]);
    } else if (!strcmp(argv[i], "-E")) {
      i++;
      p_flags |= PREPARE_USE_CPP;
    } else if (!strcmp(argv[i], "-v")) {
      p_flags &= ~PREPARE_QUIET;
    } else break;
  }
  if (i + 5 > argc) goto print_usage;

  if (sscanf(argv[i+1], "%d %n", &team, &n) != 1
      || argv[i+1][n]) goto print_usage;
  if (sscanf(argv[i+2], "%d %n", &prob, &n) != 1
      || argv[i+2][n]) goto print_usage;
  if (sscanf(argv[i+3], "%d %n", &lang, &n) != 1
      || argv[i+3][n]) goto print_usage;

  if ((p_flags & PREPARE_QUIET)) logger_set_level(-1, LOG_WARN);
  if (prepare(argv[i], p_flags, PREPARE_SERVE, cpp_opts) < 0) return 1;
  if ((p_flags & PREPARE_QUIET)) logger_set_level(-1, LOG_JUNK);

  if (teamdb_lookup(team) < 0) goto print_usage;
  if (prob <= 0 || prob > max_prob || !probs[prob]) goto print_usage;
  if (lang <= 0 || lang > max_lang || !langs[lang]) goto print_usage;

  get_uniq_prefix(uniq);
  if (generic_copy_file(0, 0, argv[i+4], "",
                        0, global->team_data_dir, uniq, "") < 0) return 1;
  sprintf(cmd, "SUBMIT %d %d %d N/A\n", team, prob, lang);
  if (generic_write_file(cmd, strlen(cmd), SAFE,
                         global->team_cmd_dir, uniq, "") < 0)
    return 1;

  return 0;

 print_usage:
  printf("  Usage: %s [ OPTS ] <cfg> <teamid> <problem> <language> <source>\n",
         argv[0]);
  printf("  -E     - enable C preprocessor\n");
  printf("  -DDEF  - define a symbol for preprocessor\n");
  printf("  -v     - verbose operation\n");
  return code;
}

/**
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE")
 * End:
 */

